"""WebSocket client class."""
import asyncio
import itertools
import json
import logging
import threading
from typing import Any, Dict, Optional

import websockets

from modelon.impact.client.sal.uri import URI
from modelon.impact.client.sal.ws_response import JsonRpcResponse

logger = logging.getLogger(__name__)


class WebSocketRPCClient:
    def __init__(self, uri: URI, api_key: Optional[str] = None):
        self._uri = (uri / "/api/modeling/rpc").resolve()
        self._api_key = api_key
        self._msg_id_gen = itertools.count()
        self._pending: Dict[int, asyncio.Future] = {}
        self._conn = None
        self._session_id = None
        self._listener_task: Optional[asyncio.Task] = None

    @property
    def session_id(self) -> str:
        if not self._session_id:
            logger.error("WebSocket connection is not established.")
            return ""
        return str(self._session_id)

    async def connect(self) -> None:
        headers = {"User-Agent": "impact-python-client"}
        if self._api_key:
            headers["impact-api-key"] = self._api_key

        self._conn = await websockets.connect(
            self._uri,
            additional_headers=headers,
        )  # type: ignore
        self._session_id = self._conn.id  # type: ignore
        logger.info(f"Connected to {self._uri}")
        self._listener_task = asyncio.create_task(self._listen())

    async def close(self) -> None:
        if self._conn:
            await self._conn.close()
        if self._listener_task:
            self._listener_task.cancel()
        logger.info("WebSocket connection closed")

    async def _listen(self) -> None:
        if not self._conn:
            logger.error("WebSocket connection is not established.")
            return
        try:
            async for message in self._conn:
                logger.debug(f"Received message: {message}")
                resp = json.loads(message)
                if resp.get("method") == "impact/workspace":
                    logger.info(resp.get("params", {}).get("message"))
                    continue
                response = JsonRpcResponse(resp)
                msg_id = response.id
                future = self._pending.pop(msg_id, None)
                if future:
                    future.set_result(response)
        except asyncio.CancelledError:
            logger.info("WebSocket listener cancelled")
        except Exception as e:
            logger.exception(f"Error in WebSocket listener. Cause {e}")

    async def send_rpc(
        self, method: str, params: Any, timeout: Optional[float] = 10.0
    ) -> Any:
        if not self._conn:
            raise RuntimeError("WebSocket is not connected")

        msg_id = next(self._msg_id_gen)
        request = {"jsonrpc": "2.0", "id": msg_id, "method": method, "params": params}
        future = asyncio.get_event_loop().create_future()
        self._pending[msg_id] = future
        await self._conn.send(json.dumps(request))
        logger.debug(f"Sent message: {request}")
        try:
            return await asyncio.wait_for(future, timeout)
        except asyncio.TimeoutError:
            self._pending.pop(msg_id, None)
            raise TimeoutError(f"Timeout waiting for response to msg_id {msg_id}")


class SyncWebSocketClient:
    def __init__(self, websocket_rpc_client: WebSocketRPCClient):
        self._loop = asyncio.new_event_loop()
        self._thread = threading.Thread(target=self._start_loop, daemon=True)
        self._thread.start()
        self._client = websocket_rpc_client
        self._run(self._client.connect())

    def _start_loop(self) -> None:
        asyncio.set_event_loop(self._loop)
        self._loop.run_forever()

    def _run(self, coro: Any) -> Any:
        return asyncio.run_coroutine_threadsafe(coro, self._loop).result()

    @property
    def session_id(self) -> str:
        return self._client.session_id

    def close(self) -> None:
        self._run(self._client.close())
        self._loop.call_soon_threadsafe(self._loop.stop)
        self._thread.join()

    def get_json_response(
        self,
        method: str,
        params: Any = None,
    ) -> Any:
        response = self._run(self._client.send_rpc(method, params, timeout=10))
        return response.result



from typing import List, Optional

from modelon.impact.client.configuration import Experimental
from modelon.impact.client.entities.model import Model
from modelon.impact.client.entities.package import Package
from modelon.impact.client.options import CompilerOptions, RuntimeOptions
from modelon.impact.client.sal.modeling import ModelingService
from modelon.impact.client.sal.service import Service


class ModelingSession:
    """Class containing ModelingService functionalities."""

    def __init__(
        self,
        session_id: str,
        workspace_id: str,
        modeling_sal: ModelingService,
        service: Service,
    ):
        self._session_id = session_id
        self._workspace_id = workspace_id
        self._modeling_sal = modeling_sal
        self._sal = service

    def __eq__(self, obj: object) -> bool:
        return isinstance(obj, ModelingSession) and obj._session_id == self._session_id

    def __repr__(self) -> str:
        return f"Modeling session with ID '{self._session_id}'"

    def close(self) -> None:
        return self._modeling_sal.close_session()

    @Experimental
    def get_available_runtime_options(self) -> RuntimeOptions:
        return RuntimeOptions(self._modeling_sal.get_available_runtime_options(), "")

    @Experimental
    def get_available_compiler_options(self) -> CompilerOptions:
        return CompilerOptions(self._modeling_sal.get_available_compiler_options(), "")

    @Experimental
    def get_packages(self) -> List[Package]:
        libraries = self._modeling_sal.get_top_classes()
        packages = []
        for project_id, top_classes in libraries.items():
            for top_class in top_classes:
                packages.append(
                    Package(
                        top_class["name"],
                        self._workspace_id,
                        project_id,
                        self._sal,
                        self._modeling_sal,
                    )
                )
        return packages

    @Experimental
    def get_experiment_definitions(self, class_path: str) -> dict:
        project_ids = [prj.get("id") for prj in self._modeling_sal.get_projects()]
        return self._modeling_sal.get_experiment_definitions(project_ids, class_path)

    @Experimental
    def find_usage(self, modelica_path: str) -> List[str]:
        return self._modeling_sal.find_usage(modelica_path)

    @Experimental
    def get_model_parmeters(self, modelica_path: str) -> List[str]:
        return self._modeling_sal.get_model_parmeters(modelica_path)

    @Experimental
    def create_top_level_package(
        self, project_id: str, package_name: str, libraryToExtend: Optional[str] = None
    ) -> Package:
        self._modeling_sal.create_top_level_package(
            project_id, package_name, libraryToExtend
        )
        return Package(
            package_name,
            self._workspace_id,
            project_id,
            self._sal,
            self._modeling_sal,
        )

    @Experimental
    def duplicate_modelica_model(
        self,
        source_modelica_path: str,
        target_modelica_path: str,
    ) -> Model:
        source_modelica_path_split = source_modelica_path.split(".")
        target_modelica_path_split = target_modelica_path.split(".")
        source_library_name = source_modelica_path_split[0]
        target_library_name = target_modelica_path_split[0]
        top_level_packages = self.get_packages()
        package_name_prj_id_map = {
            package.name: package.project_id
            for package in top_level_packages  # type: ignore
        }
        if source_library_name not in package_name_prj_id_map:
            raise ValueError(f"Source library {source_library_name} doesn't exist")
        if target_library_name not in package_name_prj_id_map:
            raise ValueError(f"Target library {target_library_name} doesn't exist")

        source_project_id = package_name_prj_id_map[source_library_name]
        target_project_id = package_name_prj_id_map[target_library_name]
        result = self._modeling_sal.duplicate_modelica_model(
            source_project_id=source_project_id,
            old_name=source_modelica_path,
            target_project_id=target_project_id,
            new_name=target_modelica_path_split[-1],
            target_package_name=target_modelica_path_split[-2],
        )
        return Model(
            result["newClassName"], self._workspace_id, target_project_id, self._sal
        )

    @Experimental
    def get_analysis_view(self, project_id: str, class_name: str) -> List[str]:
        return self._modeling_sal.get_analysis_view(project_id, class_name)

    @Experimental
    def get_model_source(self, class_path: str) -> str:
        return self._modeling_sal.get_model_source(class_path)



from typing import Any, Dict, Optional


class JsonRpcError(Exception):
    def __init__(self, message: str, code: int, data: Optional[Any] = None):
        self.message = message
        self.code = code
        self.data = data
        super().__init__(f"[{code}] {message}")


class JsonRpcResponse:
    def __init__(self, resp_obj: Dict[str, Any]):
        if "jsonrpc" not in resp_obj or resp_obj["jsonrpc"] != "2.0":
            raise ValueError(f"Invalid JSON-RPC response: {resp_obj}")
        self._resp_obj = resp_obj

    @property
    def id(self) -> int:
        msg_id = self._resp_obj.get("id")
        if msg_id is None:
            raise ValueError(
                f"Expected 'msg_id' field in JSON-RPC response: {self._resp_obj}"
            )
        return msg_id

    @property
    def ok(self) -> bool:
        return "result" in self._resp_obj

    @property
    def result(self) -> Any:
        if not self.ok:
            error = self.error
            if error is None:
                raise ValueError("Expected 'error' field in JSON-RPC error response")
            raise JsonRpcError(
                message=error.get("message", "Unknown error"),
                code=error.get("code", -32000),
                data=error.get("data"),
            )
        return self._resp_obj["result"]

    @property
    def error(self) -> Optional[Dict[str, Any]]:
        err = self._resp_obj.get("error")
        if err is None:
            raise ValueError("No error in response")
        return err
